#!/usr/bin/python3
# coding: utf8

# The Incredible Machine (TIM) level/image resource reader, version 0.2
# By knt47, licensed under the Creative Commons Zero license

from __future__ import division, print_function, unicode_literals

import itertools, sys

if sys.version_info[0] != 3:
    raise NotImplementedError('Only Python 3 is supported.')
else:
    izip = zip
    xrange = range

from collections import namedtuple
from timgres.base.misc import (print_msg, print_traceback, s_vdata_palimg,
                               s_opt_transpcol16, s_opt_transpcol256,
                               s_opt_pal_file_def, s_opt_pal_file_def_2,
                               s_opt_zoom, s_obj_pngexp16, s_obj_pngexp256,
                               namedtuple_to_str, color_dist, color_black,
                               color_white)
from timgres.base.bmpdbase import (nt_bm_data, nt_bm_sub_img_data, nt_pal_data,
                                   flip_none)
from timgres.base.vdobase import (ResDataBase_ViewerDataObj, nt_size,
                                  nt_anim_info, nt_image_display_data,
                                  nt_info_col_header)
from timgres.base.pngexp import SimplePngExporter, PNG_COL_TYPE_INDEXED

# --- TreeDataBase_ViewerDataObj subclass for palettes (.PAL) ---

'''
Module containing the "viewer data object" for palettes, ViewerDataObj_Pal.
As palette decoders return only the color information, the PNG image must
    be generated by the exporter (this is different form bitmap viewers), see
    class PaletteImageGenerator.
The module also contains the callback function for automatically loading the
    default palettes from an opened resource file.
'''

# NOTE: This is different from tn_bmp as the decoder object does not generate
#    bitmap images in this case, just returns the palette entries and some info

# Source:
# https://github.com/spacerace/romfont/blob/master/other_sources/bochs_bios/bochs8x8.h
# original description:
# * This is the 8x8 font used by BOCHS BIOS
# *
# * This font comes from the fntcol16.zip package (c) by  Joseph Gil
# * found at ftp://ftp.simtel.net/pub/simtelnet/msdos/screen/fntcol16.zip
# * This font is public domain

# the following characters were extracted: 0-9, A-F hex, upper case letters)

hex_chars_font8x8 = (
    (0x7c, 0xc6, 0xce, 0xde, 0xf6, 0xe6, 0x7c, 0x00),
    (0x30, 0x70, 0x30, 0x30, 0x30, 0x30, 0xfc, 0x00),
    (0x78, 0xcc, 0x0c, 0x38, 0x60, 0xcc, 0xfc, 0x00),
    (0x78, 0xcc, 0x0c, 0x38, 0x0c, 0xcc, 0x78, 0x00),
    (0x1c, 0x3c, 0x6c, 0xcc, 0xfe, 0x0c, 0x1e, 0x00),
    (0xfc, 0xc0, 0xf8, 0x0c, 0x0c, 0xcc, 0x78, 0x00),
    (0x38, 0x60, 0xc0, 0xf8, 0xcc, 0xcc, 0x78, 0x00),
    (0xfc, 0xcc, 0x0c, 0x18, 0x30, 0x30, 0x30, 0x00),
    (0x78, 0xcc, 0xcc, 0x78, 0xcc, 0xcc, 0x78, 0x00),
    (0x78, 0xcc, 0xcc, 0x7c, 0x0c, 0x18, 0x70, 0x00),
    (0x30, 0x78, 0xcc, 0xcc, 0xfc, 0xcc, 0xcc, 0x00),
    (0xfc, 0x66, 0x66, 0x7c, 0x66, 0x66, 0xfc, 0x00),
    (0x3c, 0x66, 0xc0, 0xc0, 0xc0, 0x66, 0x3c, 0x00),
    (0xf8, 0x6c, 0x66, 0x66, 0x66, 0x6c, 0xf8, 0x00),
    (0xfe, 0x62, 0x68, 0x78, 0x68, 0x62, 0xfe, 0x00),
    (0xfe, 0x62, 0x68, 0x78, 0x68, 0x60, 0xf0, 0x00),
)

_bpp_8 = 8  # 8 bits per pixel, max. 256 colors
_font_pixels = 8

# nt_bm_info_pal: Stored as the sub_img_info member of the
#     nt_bm_sub_img_data namedtuple
# img_w, img_h (int): The width and height of the image.
# n_colors (int): the number of colors in the palette.
nt_bm_info_pal = namedtuple('nt_bm_info_pal', ('img_w', 'img_h', 'n_colors'))

# color_dist: moved to base.misc !


def _font8_row_to_8bpp(row_byte, color_pal_entry, color_text, zoom):
    '''
    "Helper" function. Converts one byte (one row) from 8x8 font data
        to pixel data in a 8BPP image. Returns bytearray,
    row_byte (int): One byte (one row) from the font data.
    color_pal_entry (int): the color index in the palette.
    color_text (int): color indef ox the text
    zoom (int): An integer number, specifying the zoom.
    '''
    # row_byte, color_pal_entry, colot_text: must be a single byte (int)
    color_pe_zoom = bytearray((color_pal_entry, )) * zoom
    color_t_zoom = bytearray((color_text, )) * zoom

    return bytearray().join(
        color_t_zoom if row_byte & (1 << (7 - x)) else color_pe_zoom
        for x in xrange(_font_pixels))


class PaletteImageGenerator(object):
    '''
    Class for generating indexed-color PNG images from palettes, used by
        ViewerDataObj_Pal. generate_palette_image() must be called for
        generating the image. It can be retrieved using get_data().
    '''

    # NOTE this object has its own PNG exporter, which is fully independent
    #    from the png exporter stored in the ResDataMgr and its palette.
    #    This is possible as this class does not depend on the "pre-loaded"
    #    palettes of pngexp16 and pngexp256. Always generates 256-color images

    num_columns = 16
    cell_width_no_zoom = 20
    cell_height_no_zoom = 12
    cell_margins = 2
    color_bw_limit = 30000  # 128 * 128 * 3

    def __init__(self):
        '''The constructor of the class. Has no arguments.'''
        self.png_exp_pal = SimplePngExporter()
        self.png_exp_pal.set_bit_depth_color_type(_bpp_8, PNG_COL_TYPE_INDEXED)
        self.reset_data()

    def reset_data(self):
        '''
        Resets the fields containing the output data, as well as
            temporary data used during image processing.
        '''
        # out_image_data: a nt_bm_sub_img_data (base.bmpdbase) namedtuple
        #     containing the output image and data (or Nome)
        self.out_image_data = None
        # List of tuples or bytearrays (format: (R, G, B)) for each color
        self.pal_colors = None
        # zoom: The zoom (an option of a ResDataManager)
        self.zoom = None
        # color_id_black, color_id_white: The index of the "darkest" and the
        #     "brightest" color in the palette, used for text.
        self.color_id_black = None
        self.color_id_white = None

    def reset_image(self):
        '''Calls the reset_image() method of the PNG exporter.'''
        self.png_exp_pal.reset_image()

    def find_black_white(self):
        '''
        Finds the closest color in the palette to black and white.
        Sets the self.color_id_black and self.color_id_white fields.
        '''
        diff_min_black = 99999  # a large number, result must be less than this
        diff_min_white = 99999
        # NOTE: must reset only self.color_id_black  and _id_white!

        for color_i, color in enumerate(self.pal_colors):
            diff_black = color_dist(color, color_black)

            if diff_black < diff_min_black:
                self.color_id_black = color_i
                diff_min_black = diff_black

            # HACK, color #0 is often transparent, try to use a different
            #    color for black if possible

            if (diff_black == 0) and (self.color_id_black == 0):
                self.color_id_black = color_i

            diff_white = color_dist(color, color_white)

            if diff_white < diff_min_white:
                self.color_id_white = color_i
                diff_min_white = diff_white

            if (diff_white == 0) and (self.color_id_white == 0):
                self.color_id_white = color_i

        # this should never happen
        if (self.color_id_black is None) or (self.color_id_white is None):
            raise RuntimeError('find_black_white() failed')

    def pal_img_add_row(self, row_num):
        '''
        For internal use. Adds one row of palette entries to the output image.
        row_num (int): The number of row (0-15)
        '''
        n_colors_pal = len(self.pal_colors)
        cell_w_zoom = self.cell_width_no_zoom * self.zoom

        for line_num in xrange(-2, self.cell_height_no_zoom - 2):
            out_row = bytearray()

            for col_num in xrange(self.num_columns):
                color_i = (row_num << 4) | col_num

                if color_i == n_colors_pal:
                    n_blank_cells = ((row_num + 1) << 4) - n_colors_pal

                    out_row.extend(
                        bytearray((self.color_id_white, )) *
                        (n_blank_cells * cell_w_zoom))

                    break

                if 0 <= line_num < _font_pixels:
                    # does not use _color_dist
                    color_d_black = sum(x**2 for x in self.pal_colors[color_i])

                    color_text = (self.color_id_white
                                  if color_d_black < self.color_bw_limit else
                                  self.color_id_black)

                    border_2 = (bytearray(
                        (color_i, )) * (self.cell_margins * self.zoom))

                    out_row.extend(border_2)

                    out_row.extend(
                        _font8_row_to_8bpp(
                            hex_chars_font8x8[row_num][line_num], color_i,
                            color_text, self.zoom))

                    out_row.extend(
                        _font8_row_to_8bpp(
                            hex_chars_font8x8[col_num][line_num], color_i,
                            color_text, self.zoom))

                    out_row.extend(border_2)
                else:
                    out_row.extend(bytearray((color_i, )) * cell_w_zoom)

            for cnt in xrange(self.zoom):
                self.png_exp_pal.add_line(out_row)

    def generate_palette_image(self, pal_colors, zoom):
        '''
        "Main" method for generating a palette image.
        pal_colors: list of (R, G, B) tuples (or bytearrays) for each
            color in the palette.
        zoom (int): The zoom.
        '''

        self.reset_data()
        self.reset_image()

        self.pal_colors = pal_colors
        self.zoom = zoom

        img_width = self.cell_width_no_zoom * self.num_columns * zoom
        # FIXME, assumes num_columns == 16
        pal_rows = (len(pal_colors) + 15) >> 4
        img_height = pal_rows * self.cell_height_no_zoom * zoom
        self.png_exp_pal.set_width_height(img_width, img_height)
        self.png_exp_pal.set_palette(pal_colors)

        self.find_black_white()

        for cnt in xrange(pal_rows):
            self.pal_img_add_row(cnt)

        # FIXME, img_width, img_height stored twice
        bm_info_pal = nt_bm_info_pal(img_width, img_height, len(pal_colors))

        self.out_image_data = nt_bm_sub_img_data(
            bm_info_pal, img_width, img_height,
            self.png_exp_pal.get_png_data())

        self.reset_image()

    def get_data(self):
        '''
        Returns the generated image and associated information
            (type: nt_bm_sub_img_data)
        '''
        return self.out_image_data


# nt_sub_item_info_pal: mamedtuple for sub-item information (sub_item_info),
#     used in class ViewerDataObj_Pal. All members are strings.
# color_num_or_info: Text of the first column in the "information table".
# r, g, b: Information for each color of the palette
# notes: Additional information.

nt_sub_item_info_pal = namedtuple(
    'nt_sub_item_info_pal', ('color_num_or_info', 'r', 'g', 'b', 'notes'))

str_err_pal_data_type = ('parent.decode_sub_file() returned invalid '
                         'data type for .PAL')


class ViewerDataObj_Pal(ResDataBase_ViewerDataObj):
    '''
    Viewer data object class (ResDataBase_ViewerDataObj subclass) that creates
        PNG images from palettes using the PaletteImageGenerator class.
    '''

    # obj_identifier, obj_description: from ResDataBase_ViewerDataObj,
    #     modified
    obj_identifier = s_vdata_palimg
    obj_description = 'Palette'

    # info_col_headers: corresponds to the items of nt_sub_item_info_pal
    info_col_headers = (
        nt_info_col_header('Color Num.', 220),
        nt_info_col_header('R', 60),
        nt_info_col_header('G', 60),
        nt_info_col_header('B', 60),
        nt_info_col_header('Notes', 450),
    )

    def __init__(self):
        '''Constructor of the class. Requires no arguments.'''
        ResDataBase_ViewerDataObj.__init__(self)
        self.pal_img_gen = PaletteImageGenerator()

    def reset_data_subclass(self):
        '''
        Re-implementation of reset_data_subclass from
            ResDataBase_ViewerDataObj. Initializes class-specific fields.
        '''
        self.data_pal_img = None

    def open_sub_file(self, sub_f_name):
        '''
        Re-implementation of open_sub_file from ResDataBase_ViewerDataObj.
        See meaning of argument there. Calls the decode_sub_file method
            of parent for the specified palette. Generates the PNG image using
            self.pal_img.gen. Fills all list members of the class,
            required in a ViewerDataObj.
        '''
        self.reset_data()

        # Currently, palette decoders do not use "process_options"
        pal_data = self.parent.decode_sub_file(sub_f_name, None)

        if not isinstance(pal_data, nt_pal_data):
            raise RuntimeError(str_err_pal_data_type)

        # NOTE: this viewer does not support "flip"
        zoom = self.parent.get_option(s_opt_zoom)
        self.pal_img_gen.generate_palette_image(pal_data.pal_colors, zoom)

        # TODO: stored, but not necessary?
        self.data_pal_img = self.pal_img_gen.get_data()

        # always has a single image!

        self.sub_item_info.append(
            nt_sub_item_info_pal(
                'Palette image (%d colors) [0/1]' % len(pal_data.pal_colors),
                '', '', '', namedtuple_to_str(self.data_pal_img.sub_img_info)))

        # actually, no animation
        self.anim_info.append([nt_anim_info(0, None)])
        # image shown in "hardcoded" 0, 0 position

        # NOTE: png_data_flip means here a single image, not a tuple of
        #   the possible flips. This class does not support flip!

        self.img_data_frames.append([
            nt_image_display_data(0, 0, flip_none,
                                  self.data_pal_img.png_data_flip)
        ])

        self.recomm_canvas_size = nt_size(self.data_pal_img.out_w,
                                          self.data_pal_img.out_h)

        self.item_info_head = nt_sub_item_info_pal(
            'Palette (%d colors)' % len(pal_data.pal_colors), '', '', '',
            namedtuple_to_str(pal_data.pal_info))

        self.item_info_extra = []

        for cnt, (r, g, b) in enumerate(pal_data.pal_colors):
            self.item_info_extra.append(
                nt_sub_item_info_pal('Color %d' % cnt, '%d' % r, '%d' % g,
                                     '%d' % b, ''))

        # NOTE: transp_col (guess for transparent color) is NOT shown


def _open_res_calback_pal_open_single(sub_f_name, res_data_mgr, obj_pngexp16,
                                      obj_pngexp256):
    '''
    "Private" function. Used by open_res_callback_pal to set the palette
        of one ot he PNG exporters.
    sub_f_name (str): The palette sub-file name
    res_data_mgr: The "resource data manager" class (ResDataManager subclass)
        that has called the callback function.
    obj_pngexp16: A SimplePngExporter (base.pngexp) object from res_data_mgr,
        using a 16-color palette.
    obj_pngexp256: A SimplePngExporter object from res_data_mgr,
        using a 256-color palette.
    '''
    if not sub_f_name.strip(): return

    # Currently, palette decoders do not use "process_options"
    try:
        pal_data = res_data_mgr.decode_sub_file(sub_f_name, None)
    except ValueError:
        # get_single_file_data raises ValueError if the file is not available
        print_msg(('Palette file "%s" not found, this '
                   'palette will not be available') % sub_f_name)
        return

    if not isinstance(pal_data, nt_pal_data):
        raise RuntimeError(str_err_pal_data_type)

    if len(pal_data.pal_colors) > 16:
        obj_pngexp256.set_palette(pal_data.pal_colors)
        obj_pngexp256.set_transp_color(pal_data.transp_col)
        res_data_mgr.set_option(s_opt_transpcol256, pal_data.transp_col)
    else:
        obj_pngexp16.set_palette(pal_data.pal_colors)
        obj_pngexp16.set_transp_color(pal_data.transp_col)
        res_data_mgr.set_option(s_opt_transpcol16, pal_data.transp_col)


def open_res_callback_pal(res_data_mgr):
    '''
    A callback function, that can be used by a ResDataManager object
        to set the palette of the PNG exporters. Palette file names are
        stored as options in the class.
    res_data_mgr: The "resource data manager" class (ResDataManager subclass)
        that has called the callback function.
    '''
    res_data_mgr.set_option(s_opt_transpcol16, None)
    res_data_mgr.set_option(s_opt_transpcol256, None)

    obj_pngexp16 = res_data_mgr.get_object(s_obj_pngexp16)
    obj_pngexp16.reset_image()
    obj_pngexp16.set_palette(None)
    obj_pngexp16.set_transp_color(None)

    obj_pngexp256 = res_data_mgr.get_object(s_obj_pngexp256)
    obj_pngexp256.reset_image()
    obj_pngexp256.set_palette(None)
    obj_pngexp256.set_transp_color(None)

    sub_f_pal_def = res_data_mgr.get_option(s_opt_pal_file_def)
    _open_res_calback_pal_open_single(sub_f_pal_def, res_data_mgr,
                                      obj_pngexp16, obj_pngexp256)

    sub_f_pal_def_2 = res_data_mgr.get_option(s_opt_pal_file_def_2)
    _open_res_calback_pal_open_single(sub_f_pal_def_2, res_data_mgr,
                                      obj_pngexp16, obj_pngexp256)
